/*----------------------------------------------------------------------*\
 * xact.c  (X Answer to Color Table editors)				*
 * X11 Red, Green, Blue color table editor program 			*
 * /UCAR/NCAR/ATD/RDP/Forrest Cook  7/28/93				*
 * from xrgb: /UCAR/NCAR/ATD/RDP/Jon Corbet, Forrest Cook  11/20/90	*
 * Version 1.1 FC Feb 4, 1995						*
 * Version 1.2								*
\*----------------------------------------------------------------------*/
/*      Copyright (C) 1990,91,92,93,94,95,96 by UCAR
 *      University Corporation for Atmospheric Research
 *      All rights reserved
 *
 * No part of this work covered by the copyrights herein may be reproduced
 * or used in any form or by any means -- graphic, electronic, or mechanical,
 * including photocopying, recording, taping, or information storage and
 * retrieval systems -- without permission of the copyright owner.
 *
 * This software and any accompanying written materials are provided "as is"
 * without warranty of any kind.  UCAR expressly disclaims all warranties of
 * any kind, either express or implied, including but not limited to the
 * implied warranties of merchantibility and fitness for a particular purpose.
 * UCAR does not indemnify any infringement of copyright, patent, or trademark
 * through use or modification of this software.  UCAR does not provide
 * maintenance or updates for its software.
 */

#include <stdio.h>
#include <fcntl.h>
#include <X11/Xatom.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Scrollbar.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>
#include <X11/Xaw/Cardinals.h>
#include "xfileio.h"

#define RED 0
#define GREEN 1
#define BLUE 2
#define MAXCVAL 65535
#define CFMLEN 514
#define STARTLEV 128	/* starting color levels for R, G, and B */
#define DEFNUMC 16	/* default number of color cells to start with */
#define MAXCOLS 128	/* max number of color cells permitted */

/*
 * Define the colortable type indices here.
 * New types should start at the end and CT_MAXN should be incremented.
 */
#define CT_ASCII 0
#define CT_CFM 1
#define CT_PS 2
#define CT_ZEB 3
#define CT_X11 4
#define CT_MAXN 4

/*
 * Function Prototypes for xact.c
 */

void sel_rgt(), sel_mid();
void cjump(), cscroll(), xact_exit(), hex_select();
void l_select(), r_select(), lff_sel(), rff_sel();
void sel_lft(), thumbs_numbs(), norml_tbl(), movl_tbl(), movr_tbl();
void swap_marks(), swap_colors(), rotl_tbl(), rotr_tbl();
void smooth_tbl(), span_tbl(), clear_tbl();
void set_type(), set_preset(), smear_tbl(), flip_tbl();
void ins_col(), rem_col(), fp_fileok();
void init_undo(), undo(), undo_save();
void mark_on(), mark_off(), star_on(), star_off(), new_sel();
void new_cmd_btn(), new_numslide();
void file_read_fn(), file_write_fn();
void SetLabelColor ();
void SetLabelNumber ();

/*
 * Lots of ugly Global vars.
 */
Widget Top, Form, Lblk; 
Widget Csel[MAXCOLS], Cbar[MAXCOLS], *Wp, *Wq, H, V;

/*
 * Color table type names
 */
char *Ct_names[] = { "ASCII", "CFM", "PS", "ZEB", "X11", "NULL" };
char Ctname[80]={"xact"};

int Cpos = 0;		/* the location of the * color postion marker */
int Mpos = 0;		/* the location of the ^ color mark marker */
int Cact = 0;
int Mact = 0;
int Botc = 0;
int Topc = 0;
int Undoc = 0;
int Undun = 0;
int Cttype = CT_ASCII;
int Cval[MAXCOLS], Ncol;

XColor Xc[MAXCOLS];
XColor Xc_sav1[MAXCOLS];
XColor Xc_sav2[MAXCOLS];
XColor Xc_tmp;

/*----------------------------------------------------------------------*/

/*
 * Build the translations from actions to functions.
 */

Cardinal Num_actions = THREE;

static XtActionsRec Xact_actions[] =
{
        { "sel_rgt",          sel_rgt        },
        { "sel_mid",          sel_mid        }
};

/*----------------------------------------------------------------------*/
/*
 * Define a bunch of fallback resources.
 */

static String fallback_resources[] =
{
	"*background:			black",
	"*foreground:			white",
	"*Form.background:		gray",
	"*Scrollbar.foreground:		black",
	"*Scrollbar.height:		256",
	"*Scrollbar.width:		30",
	"*Label.foreground:		yellow",
	"*Label.resize:			FALSE",
	"*Label.width:			30",
	"*Label.height:			24",
	"*Command.foreground:		yellow",
	"*Command.height:		24",
	"*Command.width:		25",
	"*form.defaultDistance:		1",
	"*lff.label:			<<",
	"*lff.foreground:		green",
	"*left.label:			<",
	"*left.foreground:		green",
	"*right.label:			>",
	"*right.foreground:		green",
	"*rff.label:			>>",
	"*rff.foreground:		green",
	"*undo.label:			Undo",
	"*undo.foreground:		white",
	"*undo.width:			42",
	"*swapm.label:			<>",
	"*smooth.width:			49",
	"*smooth.label:			Smooth",
	"*smear.width:			42",
	"*smear.label:			Smear",
	"*span.width:			35",
	"*span.label:			Span",
	"*swap.width:			35",
	"*swap.label:			Swap",
	"*rotlft.width:			42",
	"*rotlft.label:			Roll<",
	"*rotrgt.width:			42",
	"*rotrgt.label:			Roll>",
	"*movel.width:			42",
	"*movel.label:			Move<",
	"*mover.width:			42",
	"*mover.label:			Move>",
	"*flip.width:			35",
	"*flip.label:			Flip",
	"*norm.width:			35",
	"*norm.label:			Norm",
	"*fpform.transient: 		False",
	"*errform.transient:		False",
        "*editfn.resize:		True",
        "*editfn.width:			320",
        "*editfn.height:		18",
	"*editfn.background:		black",
        "*popex.label: 			Cancel",
        "*popex.width: 			80",
	"*popex.foreground:		red",
        "*popok.label: 			Ok",
        "*popok.width: 			30",
	"*popok.foreground:		green",
        "*errmsg.resize:                False",
        "*errmsg.width:                 320",
        "*errmsg.height:                18",
        "*errex.label:                  Confirm",
        "*errex.width:                  320",
        "*errex.foreground:             green",
	"*insert.width:			50",
	"*insert.label:			Insert",
	"*insert.foreground:		orange",
	"*remove.width:			50",
	"*remove.label:			Remove",
	"*remove.foreground:		orange",
	"*read.width:			50",
	"*read.label:			Read",
	"*read.foreground:		red",
	"*write.width:			50",
	"*write.label:			Write",
	"*write.foreground:		red",
	"*type.width:			50",
	"*type.label:			Type",
	"*type.foreground:		red",
	"*preset.width:			50",
	"*preset.label:			Preset",
	"*preset.foreground:		cyan",
	"*quit.width:			43",
	"*quit.label:			Quit",
	"*quit.foreground:		magenta",
	"*clear.width:			43",
	"*clear.label:			Clear",
	"*clear.foreground:		magenta",
	"*blank.width:			4",
	"*blank.height:			283",
	"*blank.label:			",
	"*blank.background:		gray30",
	"*csel.label:			",
	"*mix.height:			256",
	"*mix.label:			",
	"*red.background:		red",
	"*green.background:		green",
	"*blue.background:		blue",
	NULL
};

/*----------------------------------------------------------------------*\
 * define a bunch of button widgets and callback functions		*
 * start a new row at index buttons[i].nxline if nxline >= 0,		*
 * ignore if nxline == -1 end if nxline == -2				*
\*----------------------------------------------------------------------*/

static struct button
{
	Widget butn;		/* button widget */
	char *but_name;		/* name of button */
	void (*but_funk)();	/* callback function */
	int arg;		/* arg to pass to callback function */
	int nxline;		/* next line multi function var */
} Buttons[] =
{
	{NULL,	"quit",		xact_exit,	0,	-1},
	{NULL,	"clear",	clear_tbl,	0,	-1},
	{NULL,	"read",		file_popup,	0,	-1},
	{NULL,	"type",		set_type,	0,	-1},
	{NULL,	"write",	file_popup,	1,	-1},
	{NULL,	"insert",	ins_col,	0,	-1},
	{NULL,	"remove",	rem_col,	0,	-1},
	{NULL,	"preset",	set_preset,	0,	-1},
	{NULL,	"span",		span_tbl,	0,	-1},
	{NULL,	"undo",		undo,		0,	-1},
	{NULL,	"lff",		lff_sel,	0,	 0},
	{NULL,	"left",		l_select,	0,	-1},
	{NULL,	"swapm",	swap_marks,	0,	-1},
	{NULL,	"right",	r_select,	0,	-1},
	{NULL,	"rff",		rff_sel,	0,	-1},
	{NULL,	"swap",		swap_colors,	0,	-1},
	{NULL,	"flip",		flip_tbl,	0,	-1},
	{NULL,	"movel",	movl_tbl,	0,	-1},
	{NULL,	"mover",	movr_tbl,	0,	-1},
	{NULL,	"rotlft",	rotl_tbl,	0,	-1},
	{NULL,	"rotrgt",	rotr_tbl,	0,	-1},
	{NULL,	"smooth",	smooth_tbl,	0,	-1},
	{NULL,	"smear",	smear_tbl,	0,	-2}
};

/*----------------------------------------------------------------------*\
 * define some number-slider widget pairs				*
 * end if nxline == -2							*
\*----------------------------------------------------------------------*/

static struct numslide
{
	Widget number;		/* number widget */
	Widget slider;		/* slider widget */
	char *num_name;		/* name of number widget */
	char *sld_name;		/* name of slider widget */
	int arg;		/* argument to functions */
	int nxline;		/* next line multi function var */
} Numslide[] =
{
	{NULL,	NULL,	"rval",	"red",		RED,	-1},
	{NULL,	NULL,	"gval",	"green",	GREEN,	-1},
	{NULL,	NULL,	"bval",	"blue",		BLUE,	-2}
};

/*----------------------------------------------------------------------*\
 * define a black pixmap here so we don't have to include any files.	*
\*----------------------------------------------------------------------*/

#define black_width 8
#define black_height 1
static unsigned char black_bits[] = { 0xff };

/*----------------------------------------------------------------------*\
 * xact.c starts here							*
\*----------------------------------------------------------------------*/

int main (argc, argv)
int argc;
char **argv;
{
	Arg args[10];
	XtAppContext actx;
	XtTranslations seltrans;
	Pixmap Thumb_pm;
	char selbuf[120];
	int i, n, ret=0, col_width=15;
/*
 * set the number of color slots according to the command line vals.
 */
        if (argc >= 2)
	{
		ret=sscanf (argv[1], "%d", &Ncol);
		if (!ret)
		{
			printf
("Usage: xact [###]  where [###] is the number of color steps in the table.\n");
			exit (0);
		}
	}
	
	if (!ret) Ncol = DEFNUMC;

	if (Ncol < 1)
	{
		printf ("\nSorry, it doesn't make sense to edit %d colors.\n",
				Ncol);
		exit (0);
	}

	if (Ncol > MAXCOLS)
	{
		printf ("\nSorry, I max out at %d colors.\n", MAXCOLS);
		exit (0);
	}
/*
 * Set the color cell width level based on the number of colors
 * and a tricky equation which guarantees visibility on a Sun screen.
 */
	col_width = 5 + (((MAXCOLS+1) - Ncol) / 11);
/*
 * Set the starting color values, build a cval array.
 */
	for (i=0; i<Ncol; i++)
	{
		Xc[i].red = Xc[i].green = Xc[i].blue = STARTLEV<<8;
		Xc[i].flags = DoRed | DoGreen | DoBlue;
		Cval[i] = i;
	}
/*
 * Create an application shell, and a form to go inside it.
 */
	Top = XtAppInitialize (&actx, "xact", (XrmOptionDescList)NULL, ZERO,
			&argc, argv, fallback_resources, (ArgList)NULL, ZERO);

	Form = XtCreateManagedWidget ("form",formWidgetClass,Top,NULL,ZERO);
/*
 * init the file i/o routines.
 */
	file_setup (&actx, file_read_fn, file_write_fn);
	set_df_fn ("", "xact_out", "asc");

/*
 * Add translation actions.
 */
        XtAppAddActions (actx, Xact_actions, Num_actions);

/*
 * Loop through the button structure and create the buttons and callbacks.
 */
	H=NULL;
	V=NULL;
	i=-1;

	do
	{
		i++;
		if (Buttons[i].nxline >= 0)
		{
			H = NULL;
			V = Buttons[Buttons[i].nxline].butn;
		}
		new_cmd_btn (&Buttons[i]);
	}
	while (Buttons[i].nxline > -2);

/*
 * Loop through the number-slider structure and create the color scrollbars
 */
	H=NULL;
	i=-1;

	do
	{
		i++;
		V=Buttons[10].butn;
		new_numslide (&Numslide[i]);
	}
	while (Numslide[i].nxline > -2);

/*
 * Create a blank label widget to separate the sliders from the colors.
 */
	V=Buttons[10].butn;
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, H);		n++;
	XtSetArg (args[n], XtNfromVert, V);		n++;
	Lblk = XtCreateManagedWidget ("blank", labelWidgetClass, Form,args,n);

	Wp = &Lblk;
	Wq = &Lblk;
/*
 * Build a bunch of stuff for each color position.
 */
	for (i=0; i<Ncol; i++)
	{
/*
 * Get one color cell per color label.
 */
		XAllocColorCells (XtDisplay (Top),
			DefaultColormap(XtDisplay(Top), 0),
			False, NULL, 0, &Xc[i].pixel, 1);
/*
 * Create a command widget for picking the color.
 */
		n = 0;
		XtSetArg (args[n], XtNfromHoriz, *Wp);		n++;
		XtSetArg (args[n], XtNfromVert, V);		n++;
		XtSetArg (args[n], XtNwidth, col_width);	n++;
		Csel[i] = XtCreateManagedWidget ("csel",commandWidgetClass,
						Form,args,n);
/*
 * Put together the translations for the middle and right buttons,
 * pass in the color position.
 */
		sprintf(selbuf,"<Btn2Up>: sel_mid(%d)\n <Btn3Up>: sel_rgt(%d)",
				Cval[i], Cval[i]);
		seltrans = XtParseTranslationTable (selbuf);
		XtOverrideTranslations (Csel[i], seltrans);
/*
 * Add the callback for the left button.
 */
		XtAddCallback (Csel[i], XtNcallback, sel_lft, &(Cval[i]));
		Wp = &Csel[i];
/*
 * Create a label widget to show the color.
 */
		n = 0;
		XtSetArg (args[n], XtNfromHoriz, *Wq);		n++;
		XtSetArg (args[n], XtNfromVert, *Wp);		n++;
		XtSetArg (args[n], XtNbackground, Xc[i].pixel); n++;
		XtSetArg (args[n], XtNwidth, col_width);	n++;
		Cbar[i] = XtCreateManagedWidget ("mix", labelWidgetClass,
						Form,args,n);
		Wq = &Cbar[i];
	}
/*
 * Display the top widget.
 */
	XtRealizeWidget (Top);
/*
 * Make a black bitmap for the scrollbar thumbs.
 */
	Thumb_pm = XCreateBitmapFromData (XtDisplay (Top),
		XtWindow (Numslide[RED].slider), black_bits, black_width,
				black_height);
	if (! Thumb_pm)
		printf ("\nError, can't make thumb pixmap\n");

	XtSetArg (args[0], XtNthumb, Thumb_pm);
	XtSetValues (Numslide[RED].slider, args, 1);
	XtSetValues (Numslide[GREEN].slider, args, 1);
	XtSetValues (Numslide[BLUE].slider, args, 1);
/*
 * Initialize the state of the undo table.
 */
	init_undo ();
/*
 * Initialize the state of various widgets.
 */
	clear_tbl ();
/*
 * Here's where it all happens. Exciting, huh?
 */
	XtAppMainLoop (actx);

	exit (0);
}

/*----------------------------------------------------------------------*/
/*
 * Make a generic X command widget button
 */
void new_cmd_btn (butst)
struct button *butst;
{
	int n=0;
	Arg args[3];

	XtSetArg (args[n], XtNfromHoriz, H);		n++;
	XtSetArg (args[n], XtNfromVert, V);		n++;
	H=butst->butn=XtCreateManagedWidget (butst->but_name,
					commandWidgetClass,Form,args,n);

	XtAddCallback (butst->butn,XtNcallback,butst->but_funk,&(butst->arg));
}

/*----------------------------------------------------------------------*/
/*
 * Make a generic X number slider widget pair
 */
void new_numslide (numslst)
struct numslide *numslst;
{
	int n=0;
	Arg args[3];

	XtSetArg (args[n], XtNfromHoriz, H);		n++;
	XtSetArg (args[n], XtNfromVert, V);		n++;
	V=numslst->number=XtCreateManagedWidget (numslst->num_name,
					labelWidgetClass, Form, args, n);
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, H);		n++;
	XtSetArg (args[n], XtNfromVert, V);		n++;
	XtSetArg (args[n], XtNorientation, XtorientVertical); n++;
	H=numslst->slider=XtCreateManagedWidget (numslst->sld_name,
					scrollbarWidgetClass,Form,args,n);
	XtAddCallback (Numslide[numslst->arg].slider, XtNscrollProc,
			cscroll, &(numslst->arg));
	XtAddCallback (Numslide[numslst->arg].slider, XtNjumpProc,
			cjump, &(numslst->arg));
}

/*----------------------------------------------------------------------*/
/*
 * Initialize the state of a lot of widgets.
 */
void clear_tbl ()
{
	Arg newargs[1];

	undo_save();

	for (Cpos=0; Cpos<Ncol; Cpos++)
	{
		XtSetArg (newargs[0], "label", "");
		XtSetValues (Csel[Cpos], newargs, 1);

		Xc[Cpos].red = Xc[Cpos].green = Xc[Cpos].blue = STARTLEV<<8;
		SetLabelColor ();
	}

	Mpos = 0;
	Mact = 0;
	Cpos = 0;
	star_on ();
	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Exit the program here.
 */
void xact_exit ()
{
	exit (0);
}

/*----------------------------------------------------------------------*/
/*
 * Set the scrollbar thumbs of the primary sliders.
 * Set the number values over the primary sliders.
 */
void thumbs_numbs ()
{
    XawScrollbarSetThumb(Numslide[RED].slider,
			1.0-(float)Xc[Cpos].red/(float)MAXCVAL,-1.0);
    XawScrollbarSetThumb(Numslide[GREEN].slider,
			1.0-(float)Xc[Cpos].green/(float)MAXCVAL,-1.0);
    XawScrollbarSetThumb(Numslide[BLUE].slider,
			1.0-(float)Xc[Cpos].blue/(float)MAXCVAL,-1.0);

    SetLabelNumber (Numslide[RED].number, Xc[Cpos].red>>8);
    SetLabelNumber (Numslide[GREEN].number, Xc[Cpos].green>>8);
    SetLabelNumber (Numslide[BLUE].number, Xc[Cpos].blue>>8);
}

/*----------------------------------------------------------------------*/
/*
 * Scroll the primary sliders incrementally,
 * activated by left and right mouse buttons.
 */
void cscroll (sb, color, position)
Widget sb;
int *color;
int position;
{
	int newc;

	if (!Undoc)	/* save color before first change */
	{
		Undoc=1;
		undo_save();
	}

	switch (*color)
	{
	   case RED:
		if (position > 0)
		{
			if (Xc[Cpos].red < 65280) Xc[Cpos].red += 256;
		}
		else if (Xc[Cpos].red > 255) Xc[Cpos].red -= 256;
		newc = Xc[Cpos].red;
		SetLabelNumber (Numslide[RED].number, newc>>8);
		break;
	   case GREEN:
		if (position > 0)
		{
			if (Xc[Cpos].green < 65280) Xc[Cpos].green += 256;
		}
		else if (Xc[Cpos].green > 255) Xc[Cpos].green -= 256;
		newc = Xc[Cpos].green;
		SetLabelNumber (Numslide[GREEN].number, newc>>8);
		break;
	   case BLUE:
		if (position > 0)
		{
			if (Xc[Cpos].blue < 65280) Xc[Cpos].blue += 256;
		}
		else if (Xc[Cpos].blue > 255) Xc[Cpos].blue -= 256;
		newc = Xc[Cpos].blue;
		SetLabelNumber (Numslide[BLUE].number, newc>>8);
		break;
	}

        XawScrollbarSetThumb (sb, 1.0 - (float)newc/(float)MAXCVAL, -1.0);

	SetLabelColor ();
}

/*----------------------------------------------------------------------*/
/*
 * Jump the primary sliders,
 * activated by the middle mouse button.
 */
void cjump (sb, color, pct)
Widget sb;
int *color;
XtPointer pct;
{
	int pos = MAXCVAL * (1 - (* (float *)pct));

	if (!Undoc)	/* save color before first change */
	{
		Undoc=1;
		undo_save();
	}

	switch (*color)
	{
	   case RED:
	   	Xc[Cpos].red = pos;
		SetLabelNumber (Numslide[RED].number, pos>>8);
		break;
	   case GREEN:
	   	Xc[Cpos].green = pos;
		SetLabelNumber (Numslide[GREEN].number, pos>>8);
		break;
	   case BLUE:
	   	Xc[Cpos].blue = pos;
		SetLabelNumber (Numslide[BLUE].number, pos>>8);
		break;
	}

	SetLabelColor ();
}

/*----------------------------------------------------------------------*/

void SetLabelColor ()
/*
 * Set the background color in our label widget.
 */
{
    XStoreColor(XtDisplay(Top),DefaultColormap(XtDisplay(Top),0),&(Xc[Cpos]));
}

/*----------------------------------------------------------------------*/
/*
 * Put a decimal number into a color value label widget.
 */
void SetLabelNumber (wi, num)
Widget wi;
int num;
{
	char ostr[10];
	Arg args[1];

	sprintf (ostr, "%02x", num);

	XtSetArg (args[0], XtNlabel, ostr);
	XtSetValues (wi, args, 1);
}

/*----------------------------------------------------------------------*/
/*
 * Move (fast forward) the select * to the far left position.
 */
void lff_sel ()
{
	int opos = Cpos;

	star_off();
	Cpos = 0;
	new_sel(opos);
	if (Cpos != opos) undo_save();

}

/*----------------------------------------------------------------------*/
/*
 * Move the select * left one position.
 */
void l_select (w, closure, callData)
Widget w;
XtPointer closure, callData;
{
	int opos = Cpos;

	if (Cpos > 0)
	{
		star_off();
		Cpos--;
		new_sel(opos);
		undo_save();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move the select * right one position.
 */
void r_select ()
{
	int opos = Cpos;

	if (Cpos < (Ncol-1))
	{
		star_off();
		Cpos++;
		new_sel(opos);
		undo_save();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move (fast forward) the select * to the far right position.
 */
void rff_sel ()
{
	int opos = Cpos;

	star_off();
	Cpos = Ncol-1;
	new_sel(opos);
	if (Cpos != opos) undo_save();
}

/*----------------------------------------------------------------------*/
/*
 * Swap the positions of the * and ^ marks.
 */
void swap_marks ()
{
	int tmp;

	if (Mact && (Mpos != Cpos))
	{
		star_off();
		mark_off();
		tmp = Mpos;
		Mpos = Cpos;
		Cpos = tmp;
		star_on();
		mark_on();
		thumbs_numbs ();
		undo_save();
	}
}

/*----------------------------------------------------------------------*/
/*
 * The left mouse button was pushed on selector, put a * on it.
 */
void sel_lft (w, position, dummy)
Widget w;
int *position, dummy;
{
	int opos = Cpos;

	if (*position != Cpos)
	{
		star_off();
		/*
		 * If no "^" is selected, leave one where the "*" was.
		 */
		if (!Mact)
		{
			Mpos = Cpos;
			mark_on();
		}
		Cpos = *position;
		new_sel (opos);
		undo_save();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Turn on the new selection, update thumbs and labels.
 */
void new_sel (opos)
int opos;
{
/*
 * If the old * was on top of a ^, put the ^ back.
 */
	if ((Mact) && (Mpos == opos)) mark_on();
/*
 * Turn the * on at the new location.
 */
	star_on();
/*
 * Update the RGB slider thumbs and label numbers.
 */
	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * The right mouse button was pushed on selector, put a ^ on it.
 */
void sel_rgt(w, event, params, num_params)
Widget w;
XEvent *event;
String *params;
Cardinal *num_params;
{
	int npos;

	sscanf (*params, "%d", &npos);
/*
 * Turn off the old mark if it's on.
 */
	if (Mact) mark_off();
	else Mact = 1;

	Mpos=npos;
	mark_on();
}

/*----------------------------------------------------------------------*/
/*
 * Turn off the old * mark at position Cpos
 */
void star_off ()
{
	Arg newargs[1];

	XtSetArg (newargs[0], "label", "");
	XtSetValues (Csel[Cpos], newargs, 1);
	Cact = 0;
}

/*----------------------------------------------------------------------*/
/*
 * Turn on a * mark at position Cpos.
 */
void star_on ()
{
	Arg newargs[1];

	XtSetArg (newargs[0], "label", "*");
	XtSetValues (Csel[Cpos], newargs, 1);
	Cact = 1;
}

/*----------------------------------------------------------------------*/
/*
 * Turn off the old mark if there is an active one.
 */
void mark_off ()
{
	Arg newargs[1];

	if (Mact)
	{
		if (Mpos != Cpos)
		{
			XtSetArg (newargs[0], "label", "");
			XtSetValues (Csel[Mpos], newargs, 1);
		}
		Mact = 0;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Turn on a mark at position Mpos.
 */
void mark_on ()
{
	Arg newargs[1];

	if (Mpos != Cpos)
	{
		XtSetArg (newargs[0], "label", "^");
		XtSetValues (Csel[Mpos], newargs, 1);
	}
	Mact = 1;
}

/*----------------------------------------------------------------------*/
/*
 * Check for existence of both * and ^ marks, set Bot and Top if ok.
 * Returns 0 for not active, 1 for marks ok, 2 for overlapping marks.
 */
int mark_chk ()
{
	int ret=0;

	if (Mact)
	{
		if (Cpos > Mpos)
		{
			Botc = Mpos;
			Topc = Cpos;
			ret=1;
		}
		else if (Cpos < Mpos)
		{
			Botc = Cpos;
			Topc = Mpos;
			ret=1;
		}
		else
		{
			Botc = Topc = Cpos;
			ret=2;
		}
	}
	return (ret);
}

/*----------------------------------------------------------------------*/
/*
 * Select a new color table type.
 */
void set_type (w, type, dummy)
Widget w;
int *type, dummy;
{
	Arg newargs[1];

	if (Cttype < CT_MAXN) Cttype++;
	else Cttype = 0;

	XtSetArg (newargs[0], "label", Ct_names[Cttype]);
	XtSetValues (w, newargs, 1);
/*
 * Set up default filename extensions.
 */
	switch (Cttype)
	{
	case CT_ASCII:
		set_df_fn ("", "xact_out", "asc");
	break;
	case CT_CFM:
		set_df_fn ("", "xact_out", "cfm");
	break;
	case CT_PS:
		set_df_fn ("", "xact_out", "ps");
	break;
	case CT_ZEB:
		set_df_fn ("", "xact_out", "zeb");
	break;
	case CT_X11:
		set_df_fn ("", "xact_out", "X11");
	break;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Do a linear color smooth from one mark to the other.
 */
void smooth_tbl ()
{
	long int rs, gs, bs;
	int i, steps, cps;

	if (mark_chk() == 1)
	{
		undo_save();

		steps = Topc - Botc;

		if (steps > 1)
		{
			rs = (Xc[Topc].red - Xc[Botc].red) / steps;
			gs = (Xc[Topc].green - Xc[Botc].green) / steps;
			bs = (Xc[Topc].blue - Xc[Botc].blue) / steps;

			cps = Cpos;

			for (i=Botc+1; i<Topc; i++)
			{
				Xc[i].red = Xc[i-1].red + rs;
				Xc[i].green = Xc[i-1].green + gs;
				Xc[i].blue = Xc[i-1].blue + bs;
				Cpos = i;
				SetLabelColor ();
			}
			Cpos = cps;
		}
	}
}

/*----------------------------------------------------------------------*/
/*
 * Do a "lowpass" color smear from one mark to the other.
 */
void smear_tbl ()
{
	int i, cps = Cpos;
	int cp, cc, cn;

	if (mark_chk() == 1)
	{
		undo_save();

		for (i=Botc+1; i<=Topc-1; i++)
		{
			cp = Xc[i-1].red >> 8;
			cc = Xc[i].red >> 8;
			cn = Xc[i+1].red >> 8;
			Xc[i].red = ((cp + (cc * 2) + cn) / 4) << 8;

			cp = Xc[i-1].green >> 8;
			cc = Xc[i].green >> 8;
			cn = Xc[i+1].green >> 8;
			Xc[i].green = ((cp + (cc * 2) + cn) / 4) << 8;

			cp = Xc[i-1].blue >> 8;
			cc = Xc[i].blue >> 8;
			cn = Xc[i+1].blue >> 8;
			Xc[i].blue = ((cp + (cc * 2) + cn) / 4) << 8;

			Cpos = i;
			SetLabelColor ();
		}
		Cpos = cps;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Take the color at Cpos and put it in all cells between Cpos and Mpos.
 */
void span_tbl ()
{
	int i, cps;

	if (Mact && (Cpos != Mpos))
	{
		undo_save();

		cps = Cpos;

		if (Cpos > Mpos)
		{
		    for (i=(Cpos-1); i>=Mpos; i--)
		    {
			Xc[i].red = Xc[Cpos].red;
			Xc[i].green = Xc[Cpos].green;
			Xc[i].blue = Xc[Cpos].blue;
			Cpos = i;
			SetLabelColor ();
		    }
		}
		else
		{
		    for (i=(Cpos+1); i<=Mpos; i++)
		    {
			Xc[i].red = Xc[Cpos].red;
			Xc[i].green = Xc[Cpos].green;
			Xc[i].blue = Xc[Cpos].blue;
			Cpos = i;
			SetLabelColor ();
		    }
		}
		Cpos = cps;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Rotate the colors spanned by the two marks to the left.
 */
void rotl_tbl ()
{
	int i, cps;
	XColor tmp;

	if (mark_chk() == 1)
	{
	    undo_save();

	    cps=Cpos;
	    tmp.red=Xc[Botc].red;
	    tmp.green=Xc[Botc].green;
	    tmp.blue=Xc[Botc].blue;

	    for (i=Botc; i<Topc; i++)
	    {
		Xc[i].red=Xc[i+1].red;
		Xc[i].green=Xc[i+1].green;
		Xc[i].blue=Xc[i+1].blue;

		Cpos=i;;
		SetLabelColor ();
	    }

	    Xc[Topc].red=tmp.red;
	    Xc[Topc].green=tmp.green;
	    Xc[Topc].blue=tmp.blue;
	    Cpos=Topc;
	    SetLabelColor ();

	    Cpos=cps;
	    thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Rotate the colors spanned by the two marks to the right.
 */
void rotr_tbl ()
{
	int i, cps;
	XColor tmp;

	if (mark_chk() == 1)
	{
	    undo_save();

	    cps=Cpos;
	    tmp.red=Xc[Topc].red;
	    tmp.green=Xc[Topc].green;
	    tmp.blue=Xc[Topc].blue;

	    for (i=Topc; i>Botc; i--)
	    {
		Xc[i].red=Xc[i-1].red;
		Xc[i].green=Xc[i-1].green;
		Xc[i].blue=Xc[i-1].blue;

		Cpos=i;;
		SetLabelColor ();
	    }

	    Xc[Botc].red=tmp.red;
	    Xc[Botc].green=tmp.green;
	    Xc[Botc].blue=tmp.blue;
	    Cpos=Botc;
	    SetLabelColor ();

	    Cpos=cps;
	    thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move the colors spanned by the two marks to the left.
 */
void movl_tbl ()
{
	int i, cps;
	XColor tmp;
	Arg newargs[1];

	if ((mark_chk() >= 1) && (Botc > 0))
	{
	    undo_save();

	    cps=Cpos;
	    tmp.red=Xc[Botc-1].red;
	    tmp.green=Xc[Botc-1].green;
	    tmp.blue=Xc[Botc-1].blue;

	    for (i=Botc-1; i<Topc; i++)
	    {
		Xc[i].red=Xc[i+1].red;
		Xc[i].green=Xc[i+1].green;
		Xc[i].blue=Xc[i+1].blue;

		Cpos=i;;
		SetLabelColor ();
	    }

	    Xc[Topc].red=tmp.red;
	    Xc[Topc].green=tmp.green;
	    Xc[Topc].blue=tmp.blue;
	    Cpos=Topc;
	    SetLabelColor ();

	    Cpos=cps;

	    XtSetArg (newargs[0], "label", "");
	    XtSetValues (Csel[Mpos], newargs, 1);
	    Mpos--;
	    XtSetArg (newargs[0], "label", "^");
	    XtSetValues (Csel[Mpos], newargs, 1);

	    XtSetArg (newargs[0], "label", "");
	    XtSetValues (Csel[Cpos], newargs, 1);
	    Cpos--;
	    XtSetArg (newargs[0], "label", "*");
	    XtSetValues (Csel[Cpos], newargs, 1);

	    thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Move the colors spanned by the two marks to the right.
 */
void movr_tbl ()
{
	int i, cps;
	XColor tmp;
	Arg newargs[1];

	if ((mark_chk() >= 1) && (Topc < (Ncol-1)))
	{
	    undo_save();

	    cps=Cpos;
	    tmp.red=Xc[Topc+1].red;
	    tmp.green=Xc[Topc+1].green;
	    tmp.blue=Xc[Topc+1].blue;

	    for (i=Topc+1; i>Botc; i--)
	    {
		Xc[i].red=Xc[i-1].red;
		Xc[i].green=Xc[i-1].green;
		Xc[i].blue=Xc[i-1].blue;

		Cpos=i;;
		SetLabelColor ();
	    }

	    Xc[Botc].red=tmp.red;
	    Xc[Botc].green=tmp.green;
	    Xc[Botc].blue=tmp.blue;
	    Cpos=Botc;
	    SetLabelColor ();

	    Cpos=cps;

	    XtSetArg (newargs[0], "label", "");
	    XtSetValues (Csel[Mpos], newargs, 1);
	    Mpos++;
	    XtSetArg (newargs[0], "label", "^");
	    XtSetValues (Csel[Mpos], newargs, 1);

	    XtSetArg (newargs[0], "label", "");
	    XtSetValues (Csel[Cpos], newargs, 1);
	    Cpos++;
	    XtSetArg (newargs[0], "label", "*");
	    XtSetValues (Csel[Cpos], newargs, 1);

	    thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Flip the colors spanned by the two marks.
 */
void flip_tbl ()
{
	int i, j, cps;
	XColor tmp;

	if (mark_chk() == 1)
	{
		undo_save();

		cps=Cpos;

		for (i=Botc,j=Topc; i<j; i++,j--)
		{
		    tmp.red=Xc[j].red;
		    tmp.green=Xc[j].green;
		    tmp.blue=Xc[j].blue;
		    Xc[j].red=Xc[i].red;
		    Xc[j].green=Xc[i].green;
		    Xc[j].blue=Xc[i].blue;
		    Xc[i].red=tmp.red;
		    Xc[i].green=tmp.green;
		    Xc[i].blue=tmp.blue;
		    Cpos=i;
		    SetLabelColor ();
		    Cpos=j;
		    SetLabelColor ();
		}
		Cpos=cps;

		thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Experimental...
 * Normalize the perceived intensity of the colors spanned by the two marks
 * using the formula: I = 0.30 * r + 0.59 * g + 0.11 * b
 */
void norml_tbl ()
{
	int i, cps;
	float norm[MAXCOLS], avint=0.0, change, cval;
	float minavi[MAXCOLS];

	if (mark_chk() == 1)
	{
		undo_save();

		cps=Cpos;
/*
 * Calculate individual intensities.
 */
		for (i=Botc; i<=Topc; i++)
		{
		    norm[i] = 0.30 * (float)(Xc[i].red) +
			      0.59 * (float)(Xc[i].green) +
			      0.11 * (float)(Xc[i].blue);
		    avint += norm[i];
		    printf ("norm[%d]=%f\n", i, norm[i]);
		}
/*
 * Calculate average intensity.
 */
		avint /= (float)(Topc-Botc+1);
		printf ("avint=%f\n", avint);
/*
 * Find dim colors who's primaries are maxed out,
 * set the average intensity to their intensity.
 */
		for (i=Botc; i<=Topc; i++)
		{
			minavi[i]=0.0;
			if ((norm[i] > 1.0) &&
			    ((Xc[i].red > 65535) ||
			     (Xc[i].green > 65535) ||
			     (Xc[i].blue > 65535)))
				minavi[i]=norm[i];
		}

		for (i=Botc; i<=Topc; i++)
			if ((minavi[i] > 0.0) && (minavi[i] < avint))
				avint = minavi[i];
/*
 * Calculate intensity change.
 */
		for (i=Botc; i<=Topc; i++)
		{
		    change = avint / norm[i];
		    printf ("change=%f\n", change);
		    cval = (unsigned int)((float)Xc[i].red * change);
		    Xc[i].red = cval;
		    printf ("cval-r:%f\n", cval);
		    cval = (unsigned int)((float)Xc[i].green * change);
		    Xc[i].green = cval;
		    printf ("cval-g:%f\n", cval);
		    cval = (unsigned int)((float)Xc[i].blue * change);
		    Xc[i].blue = cval;
		    printf ("cval-b:%f\n", cval);

		    Cpos=i;
		    SetLabelColor ();
		}
		Cpos=cps;

		thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Put one of many preset colors at the currently selected (*) position.
 */
# define NUMPRES 11

void set_preset ()
{
	static int cur_pres=NUMPRES;
	static unsigned char rpre[NUMPRES] =
	{0xff, 0xff, 0xff, 0xa9, 0x00, 0x00, 0x00, 0xff, 0x00, 0x80, 0xff};
	static unsigned char gpre[NUMPRES] =
	{0x00, 0x98, 0xff, 0xff, 0xff, 0xff, 0x00, 0x00, 0x00, 0x80, 0xff};
	static unsigned char bpre[NUMPRES] =
	{0x00, 0x00, 0x00, 0x00, 0x00, 0xff, 0xff, 0xff, 0x00, 0x80, 0xff};

	undo_save();

	if (cur_pres < (NUMPRES-1)) cur_pres++;
	else cur_pres=0;

	Xc[Cpos].red = rpre[cur_pres] << 8;
	Xc[Cpos].green = gpre[cur_pres] << 8;
	Xc[Cpos].blue = bpre[cur_pres] << 8;

	SetLabelColor ();
	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Remove the color cell at the (*) by shifting right hand colors to
 * the left and filling the right most slot with the default color.
 */
void rem_col ()
{
	int cps=Cpos;

	undo_save();
/*
 * Move all colors from right of the mark left by one.
 */
	for (Cpos=cps; Cpos<(Ncol-1); Cpos++)
	{
		Xc[Cpos].red = Xc[Cpos+1].red;
		Xc[Cpos].green = Xc[Cpos+1].green;
		Xc[Cpos].blue = Xc[Cpos+1].blue;
		SetLabelColor ();
	}
/*
 * Grey out the last color.
 */
	Cpos = Ncol-1;
	Xc[Cpos].red = STARTLEV << 8;
	Xc[Cpos].green = STARTLEV << 8;
	Xc[Cpos].blue = STARTLEV << 8;
	SetLabelColor ();

	Cpos=cps;
/*
 * Adjust the mark if necessary.
 */
	if (Mact && (Mpos > cps))
	{
		mark_off();
		Mpos--;
		mark_on();
	}

	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Insert a new color cell at the (*) by shifting colors from and to the
 * right of the (*) to the right by one position.
 */
void ins_col ()
{
	int cps=Cpos;

	undo_save();
/*
 * Move all colors from the mark right by one.
 */
	for (Cpos=Ncol-1; Cpos>cps; Cpos--)
	{
		Xc[Cpos].red = Xc[Cpos-1].red;
		Xc[Cpos].green = Xc[Cpos-1].green;
		Xc[Cpos].blue = Xc[Cpos-1].blue;
		SetLabelColor ();
	}
/*
 * Grey out the current position.
 */
	Cpos=cps;
	Xc[Cpos].red = STARTLEV << 8;
	Xc[Cpos].green = STARTLEV << 8;
	Xc[Cpos].blue = STARTLEV << 8;
	SetLabelColor ();

/*
 * Adjust the mark if necessary.
 */
	if (Mact && (Mpos > Cpos) && (Mpos < (Ncol-1)))
	{
		mark_off();
		Mpos++;
		mark_on();
	}

	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Swap the colors at the two marks.
 */
void swap_colors ()
{
	int cps=Cpos;
	XColor tmp;

	if (Mact && (Mpos != Cpos))
	{
		undo_save();

		tmp.red=Xc[Mpos].red;
		tmp.green=Xc[Mpos].green;
		tmp.blue=Xc[Mpos].blue;

		Xc[Mpos].red=Xc[Cpos].red;
		Xc[Mpos].green=Xc[Cpos].green;
		Xc[Mpos].blue=Xc[Cpos].blue;

		Cpos=Mpos;
		SetLabelColor ();
		Cpos=cps;

		Xc[Cpos].red=tmp.red;
		Xc[Cpos].green=tmp.green;
		Xc[Cpos].blue=tmp.blue;
		SetLabelColor ();

		thumbs_numbs ();
	}
}

/*----------------------------------------------------------------------*/
/*
 * Initialize the state of the undo table.
 */
void init_undo ()
{
	for (Cpos=0; Cpos<Ncol; Cpos++)
	{
		Xc_sav1[Cpos].red =
		Xc_sav1[Cpos].green =
		Xc_sav1[Cpos].blue = STARTLEV<<8;
		Xc_sav2[Cpos].red =
		Xc_sav2[Cpos].green =
		Xc_sav2[Cpos].blue = STARTLEV<<8;
	}

	Cpos = 0;
}

/*----------------------------------------------------------------------*/
/*
 * Save the table state before making a change.
 * Also save the previously saved undo state.
 */
void undo_save()
{
	int i;

	for (i=0; i<Ncol; i++)
	{
		Xc_sav2[i] = Xc_sav1[i];
		Xc_sav1[i] = Xc[i];
	}
	Undun = 0;
}

/*----------------------------------------------------------------------*/
/*
 * Undo to the previous state of the table.
 */
void undo()
{
	int cps=Cpos;

	Undoc=0;

	if (!Undun)	/* if not undone, undo */
	{
		for (Cpos=0; Cpos<Ncol; Cpos++)
		{
			Xc_sav2[Cpos] = Xc[Cpos];
			Xc[Cpos] = Xc_sav1[Cpos];
			SetLabelColor ();
		}
		Undun = 1;
	}
	else		/* IF undone, redo */
	{
		for (Cpos=0; Cpos<Ncol; Cpos++)
		{
			Xc[Cpos] = Xc_sav2[Cpos];
			SetLabelColor ();
			Xc_tmp = Xc_sav2[Cpos];	/* swap color buffer data */
			Xc_sav2[Cpos] = Xc_sav1[Cpos];
			Xc_sav1[Cpos] = Xc_tmp;
		}
		Undun = 1;
	}

	Cpos = cps;

	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Null stub for the middle mouse button.
 */
void sel_mid()
{
/*	printf ("gronk\n"); */
}

/*----------------------------------------------------------------------*/
/*
 * Read an already open colortable file using various formats.
 */
void file_read_fn (inf)
FILE *inf;
{
	int i, j, fsr, dummy, nr, nc, ncs, rv, gv, bv, ovf=0;
	float rfv, gfv, bfv;
	unsigned char chr;

	undo_save();
	clear_tbl();

	switch (Cttype)
	{
/*
 * ASCII color table input.
 */
	case CT_ASCII:
	    fscanf (inf, "(%03d)\n", &nc);

	    /*printf ("\nnc=%d", nc);*/
	    ncs=nc;
	    if (nc > Ncol)
	    {
		ovf=1;
		ncs=Ncol;
	    }

	    for (i=0; i<ncs; i++)
	    {
		fscanf (inf, "(%03d)%03d,%03d,%03d", &dummy, &rv, &gv, &bv);
		Xc[i].red = rv << 8;
		Xc[i].green = gv << 8;
		Xc[i].blue = bv << 8;

		/*printf ("\nr=%d, g=%d, b=%d", rv, gv, bv);*/

		if ((i+1)%4) fscanf (inf, " ");
		else fscanf (inf, "\n");
	    }
	break;
/*
 * CFM color table input.
 */
	case CT_CFM:
	    ncs=nc=Ncol;

	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    fread (&chr, 1, 1, inf);
		    if ((j<Ncol) && (!(i%4)))
			    Xc[j++].green = ((int)chr << 8);
		    /*printf ("\ng:%d", (int)chr);*/
	    }

	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    fread (&chr, 1, 1, inf);
		    if ((j<Ncol) && (!(i%4)))
			    Xc[j++].red = ((int)chr << 8);
		    /*printf ("\nr:%d", (int)chr);*/
	    }

	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    fread (&chr, 1, 1, inf);
		    if ((j<Ncol) && (!(i%4)))
			    Xc[j++].blue = ((int)chr << 8);
		    /*printf ("\nb:%d", (int)chr);*/
	    }
	    break;

/*
 * PostScript color table input.
 */
	case CT_PS:
	    fscanf (inf, "[/Indexed /DeviceRGB %d\n<\n", &nc);

	    /*printf ("\nnc=%d", nc);*/

	    ncs=nc;
	    if (nc > Ncol)
	    {
		ovf=1;
		ncs=Ncol;
	    }

	    for (i=0; i<ncs; i++)
	    {
		fscanf (inf, "%02x%02x%02x\n", &rv, &gv, &bv);
	    
		Xc[i].red = rv << 8;
		Xc[i].green = gv << 8;
		Xc[i].blue = bv << 8;

		/*printf ("\nr=%d, g=%d, b=%d", rv, gv, bv);*/

		fscanf (inf, "\n");
	    }
	break;
/*
 * ZEB color table input.
 */
	case CT_ZEB:
	    fscanf (inf, "colortable %s\n", Ctname);

	    /*printf ("\nfile:%s, Ctname=%s", Filename,Ctname);*/
	    nc=0;
	    do
	    {
		fsr=fscanf (inf, "\t%f\t%f\t%f\n", &rfv, &gfv, &bfv);
		if (fsr == 3)
		{
			if (!ovf)
			{
			    Xc[nc].red = rfv * (float)MAXCVAL;
			    Xc[nc].green = gfv * (float)MAXCVAL;
			    Xc[nc].blue = bfv * (float)MAXCVAL;
			}
			nc++;

			/* just ignore extra colors */
			if (nc > Ncol)
			{
				ovf=1;
				ncs=Ncol;
			}
			else ncs=nc;
		}
		/*printf ("\nfsr:%d, %f,%f,%f", fsr, rfv, gfv, bfv);*/
	    } while (fsr == 3);
	break;
/*
 * X11 color table input.
 */
	case CT_X11:
	    /*printf ("\nnc=%d", nc);*/
	    nc=0;

	    do
	    {
		nr = fscanf (inf, "rgb:%02x/%02x/%02x\n", &rv, &gv, &bv);
		Xc[nc].red = rv << 8;
		Xc[nc].green = gv << 8;
		Xc[nc].blue = bv << 8;
		nc++;

		/*printf ("\nr=%d, g=%d, b=%d", rv, gv, bv);*/

	        if ((nc > Ncol) && (nr != EOF))
	        {
		    ovf=1;
		    ncs=Ncol;
		    break;
	        }
	    }
	    while (nr != EOF);
	    ncs = nc - 1;

	break;
	}

	if (ovf) printf
    ("\n\007Sorry, only space for %d of %d colors.  Restart with: xact %d\n",
			Ncol, nc, nc);
	for (Cpos=0; Cpos<ncs; Cpos++)
		SetLabelColor ();

	Cpos = 0;

	thumbs_numbs ();
}

/*----------------------------------------------------------------------*/
/*
 * Write an already open colortable file using various formats.
 */
void file_write_fn (of)
FILE *of;
{
	int i, j;
	char obuf[CFMLEN];

	switch (Cttype)
	{
/*
 * ASCII color table output.
 */
	case CT_ASCII:
	    fprintf (of, "(%03d)\n", Ncol);

	    for (i=0; i<Ncol; i++)
	    {
		fprintf (of, "(%03d)%03d,%03d,%03d", i,
			Xc[i].red >> 8, Xc[i].green >> 8, Xc[i].blue >> 8);

		if ((i+1)%4) fprintf (of, " ");
		else fprintf (of, "\n");
	    }
	break;
/*
 * CFM color table output.
 */
	case CT_CFM:
	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    if ((j<Ncol) && (!(i%2)))
		    {
			obuf[i] = (unsigned char)(Xc[j].green >> 8);
			if (!(i%4)) j++;	/* do each color twice */
		    }
		    else obuf[i] = (unsigned char)0;
	    }
	    fwrite (obuf, 1, CFMLEN, of);

	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    if ((j<Ncol) && (!(i%2)))
		    {
			obuf[i] = (unsigned char)(Xc[j].red >> 8);
			if (!(i%4)) j++;	/* do each color twice */
		    }
		    else obuf[i] = (unsigned char)0;
	    }
	    fwrite (obuf, 1, CFMLEN, of);

	    j=0;
	    for (i=0; i<CFMLEN; i++)
	    {
		    if ((j<Ncol) && (!(i%2)))
		    {
			obuf[i] = (unsigned char)(Xc[j].blue >> 8);
			if (!(i%4)) j++;	/* do each color twice */
		    }
		    else obuf[i] = (unsigned char)0;
	    }
	    fwrite (obuf, 1, CFMLEN, of);

	break;
/*
 * PostScript color table output.
 */
	case CT_PS:
	    fprintf (of, "[/Indexed /DeviceRGB %d\n<\n", Ncol);

	    for (i=0; i<Ncol; i++) fprintf (of, "%02x%02x%02x\n",
			Xc[i].red >> 8, Xc[i].green >> 8, Xc[i].blue >> 8);

	    fprintf (of, ">\n] setcolorspace\n");
	break;
/*
 * ZEB color table output.
 */
	case CT_ZEB:
	    fprintf (of, "colortable %s\n", Ctname);

	    for (i=0; i<Ncol; i++)
	    {
		fprintf (of, "\t%0.3f\t%0.3f\t%0.3f\n",
			(float)Xc[i].red/(float)MAXCVAL,
			(float)Xc[i].green/(float)MAXCVAL,
			(float)Xc[i].blue/(float)MAXCVAL);
	    }

	    fprintf (of, "endtable\n");
	break;
/*
 * X11 color table output.
 */
	case CT_X11:
	    for (i=0; i<Ncol; i++)
	    {
		fprintf (of, "rgb:%02x/%02x/%02x\n",
			Xc[i].red >> 8, Xc[i].green >> 8, Xc[i].blue >> 8);
	    }
	break;
	}
}

/*----------------------------------------------------------------------*/
