/*----------------------------------------------------------------------*\
 * xfileio.c								*
 * G. Forrest Cook  2/6/95						*
\*----------------------------------------------------------------------*/
/*      Copyright (C) 1990,91,92,93,94,95,96 by UCAR
 *      University Corporation for Atmospheric Research
 *      All rights reserved
 *
 * No part of this work covered by the copyrights herein may be reproduced
 * or used in any form or by any means -- graphic, electronic, or mechanical,
 * including photocopying, recording, taping, or information storage and
 * retrieval systems -- without permission of the copyright owner.
 *
 * This software and any accompanying written materials are provided "as is"
 * without warranty of any kind.  UCAR expressly disclaims all warranties of
 * any kind, either express or implied, including but not limited to the
 * implied warranties of merchantibility and fitness for a particular purpose.
 * UCAR does not indemnify any infringement of copyright, patent, or trademark
 * through use or modification of this software.  UCAR does not provide
 * maintenance or updates for its software.
 */

#include <stdio.h>
#include <fcntl.h>
#include <X11/Intrinsic.h>
#include <X11/StringDefs.h>
#include <X11/Xaw/Cardinals.h>
#include <X11/Shell.h>
#include <X11/Xaw/AsciiText.h>
#include <X11/Xaw/Form.h>
#include <X11/Xaw/Label.h>
#include <X11/Xaw/Command.h>

#include "xfileio.h"

void (*Fread)();
void (*Fwrite)();

#define FNSTLEN 120		/* max filename length */

/*
 * More Global vars.
 */
extern Widget Top;
Widget Filepop;
Widget Errpop;
char Filename[FNSTLEN]={"\0"};
char Defltfnr[FNSTLEN]={"\0"};
char Defltfnw[FNSTLEN]={"\0"};
char Inpfile[FNSTLEN];
char Outfile[FNSTLEN];
char Fn_ext[10];
int File_rw = 0;

/*----------------------------------------------------------------------*/
/*
 * set up the file actions.
 */
void file_setup (act, readfn, writefn)
XtAppContext *act;
void (*readfn)();
void (*writefn)();
{
/*
 * set up an actions table for detecting cr/lf in the filename edit window.
 */
	Cardinal Num_actions = ONE;

	static XtActionsRec Xact_actions[] =
	{
		{ "fp_fileok",        fp_fileok      }
	};

        XtAppAddActions (*act, Xact_actions, Num_actions);

/*
 * Setup functions for file reads and writes.
 */
	Fread = readfn;
	Fwrite = writefn;
}

/*----------------------------------------------------------------------*/
/*
 * set new default filenames and extensions.
 */
void set_df_fn (dfr, dfw, ext)
char *dfr, *dfw, *ext;
{
	strcpy (Defltfnr, "");
	strcpy (Inpfile, dfr);
	strcpy (Defltfnw, "");
	strcpy (Outfile, dfw);
	strcpy (Fn_ext, ext);
}

/*----------------------------------------------------------------------*/
/*
 * Read in a file.
 */
void read_file ()
{
	int err=0;
	FILE *inf;
	char ostr[100];
/*
 * Call the passed in file read function if the open worked.
 */
	if ((inf = fopen(Filename, "r")) == NULL) err=1;
	else (*Fread)(inf);

	if (err)
	{
		sprintf (ostr, "File open error: %s", Filename);
		err_confirm_popup (ostr);
	}
	else fclose (inf);
}

/*----------------------------------------------------------------------*/
/*
 * Write out a file.
 */
void write_file ()
{
	int err=0;
	FILE *of;
	char ostr[100];

	if ((of = fopen(Filename, "w")) == NULL) err=1;
	else (*Fwrite)(of);

	if (err)
	{
		sprintf (ostr, "File write error: %s", Filename);
		err_confirm_popup (ostr);
	}
	else fclose (of);
}

/*----------------------------------------------------------------------*/
/*
 * Remove the filename popup without doing anything.
 */
void fp_dismiss ()
{
        XtPopdown (Filepop);
}

/*----------------------------------------------------------------------*/
/*
 * Remove the error popup box.
 */
void err_dismiss ()
{
        XtPopdown (Errpop);
}

/*----------------------------------------------------------------------*/
/*
 * Remove the filename popup and read or write a file.
 */
void fp_fileok ()
{
        XtPopdown (Filepop);

	switch (File_rw)
	{
	case 0:
		read_file();
		strcpy (Defltfnr, Filename);
	break;
	case 1:
		write_file();
		strcpy (Defltfnw, Filename);
	break;
	}
}

/*----------------------------------------------------------------------*/
/*
 * Pop up a widget for getting read/write filenames.
 */
void file_popup (w, type, dummy)
Widget w;
int *type, dummy;
{
	Widget fpform, editfn, fp_dism, fp_ok;
	Arg args[20];
	XtTranslations fntrans;
	int i, n, fn_cpos;

	File_rw = *type;

/*
 * Set the filename up for read or write.
 */
	switch (File_rw)
	{
	case 0:
		if (!strlen(Defltfnr))
			sprintf (Filename, "%s.%s", Inpfile, Fn_ext);
		else strcpy (Filename, Defltfnr);
		fn_cpos = 0;
	break;
	case 1:
		if (!strlen(Defltfnw))
			sprintf (Filename, "%s.%s", Outfile, Fn_ext);
		else strcpy (Filename, Defltfnw);
		fn_cpos = strlen(Filename);
	break;
	}
/*
 * Put the cursor on the dot.
 */
	for (i=0; i<strlen(Filename); i++)
		if (Filename[i] == '.') fn_cpos = i;
/*
 * Create a top level shell and a box for the popup.
 */
	Filepop = XtCreatePopupShell ("filename", topLevelShellWidgetClass,
					Top, (Arg *)NULL, 0);

	fpform = XtCreateManagedWidget ("fpform", formWidgetClass, Filepop,
					(Arg *)NULL, 0);
/*
 * Create an edit window for the popup box.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, NULL);		n++;
	XtSetArg (args[n], XtNfromVert, NULL);		n++;
	XtSetArg (args[n], XtNdisplayPosition, 0);      n++;
	XtSetArg (args[n], XtNinsertPosition, fn_cpos); n++;
	XtSetArg (args[n], XtNlength, FNSTLEN);		n++;
	XtSetArg (args[n], XtNtype, XawAsciiString);    n++;
        XtSetArg (args[n], XtNuseStringInPlace, True);  n++;
	XtSetArg (args[n], XtNresize, XawtextResizeWidth);      n++;
	XtSetArg (args[n], XtNstring, Filename);	n++;
	XtSetArg (args[n], XtNleftMargin, 5);           n++;
	XtSetArg (args[n], XtNeditType, XawtextEdit);   n++;
	XtSetArg (args[n], XtNdisplayCaret, TRUE);	n++;

	editfn = XtCreateManagedWidget("editfn",asciiTextWidgetClass,
					fpform, args, n);
/*
 * Put together the button translations.
 */
        fntrans = XtParseTranslationTable ("<Key>Return:   fp_fileok()\n\
					    <Key>Linefeed: fp_fileok()");
	XtOverrideTranslations (editfn, fntrans);
/*
 * Create some buttons for the popup box.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, NULL);		n++;
	XtSetArg (args[n], XtNfromVert, editfn);	n++;
	fp_ok = XtCreateManagedWidget ("popok", commandWidgetClass,
					fpform, args, n);
	XtAddCallback (fp_ok, XtNcallback, fp_fileok, NULL);

	n = 0;
	XtSetArg (args[n], XtNfromHoriz, fp_ok);	n++;
	XtSetArg (args[n], XtNfromVert, editfn);	n++;
	fp_dism = XtCreateManagedWidget ("popex", commandWidgetClass,
					fpform, args, n);
	XtAddCallback (fp_dism, XtNcallback, fp_dismiss, NULL);

	XtPopup (Filepop, XtGrabNonexclusive);
}

/*----------------------------------------------------------------------*/
/*
 * Pop up a widget for confirming file i/o errors.
 */
void err_confirm_popup (msg)
char *msg;
{
	Widget errform, err_mesg, err_dism;
	Arg args[5];
	int n;

/*
 * Create a top level shell and a box for the popup.
 */
	Errpop = XtCreatePopupShell ("Error", topLevelShellWidgetClass,
					Top, (Arg *)NULL, 0);

	errform = XtCreateManagedWidget ("errform", formWidgetClass, Errpop,
					(Arg *)NULL, 0);
/*
 * Create an edit window for the popup box.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, NULL);		n++;
	XtSetArg (args[n], XtNfromVert, NULL);		n++;
	XtSetArg (args[n], XtNlabel, msg);		n++;

	err_mesg = XtCreateManagedWidget("errmsg",labelWidgetClass,
					errform, args, n);
/*
 * Create a button for the popup box.
 */
	n = 0;
	XtSetArg (args[n], XtNfromHoriz, NULL);	n++;
	XtSetArg (args[n], XtNfromVert, err_mesg);	n++;
	err_dism = XtCreateManagedWidget ("errex", commandWidgetClass,
					errform, args, n);
	XtAddCallback (err_dism, XtNcallback, err_dismiss, NULL);

	XtPopup (Errpop, XtGrabExclusive);
}

/*----------------------------------------------------------------------*/
